/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <stdio.h>
#include <errno.h>
#include <glib.h>
#include <sys/mman.h>
#include <dlog.h>
#include <Elementary.h>
#include <Ecore_Wayland.h>
#include <wayland-extension/tizen-extension-client-protocol.h>
#include <tizen-remote-surface-client-protocol.h>
#include <tbm_surface.h>
#include <tbm_surface_internal.h>

#include "screen_connector_watcher.h"
#include "screen_connector_watcher_evas.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "SC_WATCHER_EVAS"

static Evas_Object *__viewer_win;

struct _screen_connector_watcher_evas_h {
	screen_connector_watcher_evas_ops *ops;
	screen_connector_watcher_h watcher_h;
	void *data;
};

struct _img_info_h {
	Evas_Object *img_tbm;
	Evas_Object *img_file;
	struct wl_buffer *pre_buffer;
	char *appid;
	char *instance_id;
	int pid;
};

typedef struct _img_info_h *img_info_h;
static GHashTable *__img_info_table = NULL;

static void __clear_img_tbm(img_info_h info)
{
	if (info->img_tbm == NULL)
		return;

	evas_object_del(info->img_tbm);
	info->img_tbm = NULL;
}

static void __clear_img_file(img_info_h info)
{
	if (info->img_file == NULL)
		return;

	evas_object_del(info->img_file);
	info->img_file = NULL;
}


static img_info_h __create_img_info(const char *appid, const char *instance_id, int pid)
{
	img_info_h info = (img_info_h)calloc(1, sizeof(struct _img_info_h));
	info->appid = strdup(appid);
	info->instance_id = strdup(instance_id);
	info->pid = pid;
	return info;
}

static void __destroy_img_info(gpointer data)
{
	img_info_h info = (img_info_h)data;
	if (info == NULL)
		return;
	if (info->appid) {
		free(info->appid);
		info->appid = NULL;
	}

	if (info->instance_id) {
		free(info->instance_id);
		info->instance_id = NULL;
	}
	__clear_img_file(info);
	__clear_img_tbm(info);

	free(info);
}


EXPORT_API int screen_connector_watcher_evas_init(Evas_Object *win)
{
	__viewer_win = win;
	__img_info_table = g_hash_table_new_full(g_str_hash,
			g_str_equal, free, __destroy_img_info);
	return screen_connector_watcher_init();
}

EXPORT_API int screen_connector_watcher_evas_fini(void)
{
	g_hash_table_destroy(__img_info_table);
	return screen_connector_watcher_fini();
}

static void __watcher_added_cb(const char *appid, const char *instance_id, const int pid, void *data)
{
	LOGE("__watcher_added_cb %s %s %d", appid, instance_id, pid);
	screen_connector_watcher_evas_h handle = (screen_connector_watcher_evas_h)data;

	if (handle)
		handle->ops->added_cb(appid, instance_id, pid, handle->data);
}

static void __watcher_removed_cb(const char *appid, const char *instance_id, const int pid, void *data)
{
	screen_connector_watcher_evas_h handle = (screen_connector_watcher_evas_h)data;
	img_info_h cur_info;

	if (handle)
		handle->ops->removed_cb(appid, instance_id, pid, handle->data);

	cur_info = g_hash_table_lookup(__img_info_table, instance_id);
	if (cur_info)
		g_hash_table_remove(__img_info_table, instance_id);
}

static void __watcher_update_cb(struct tizen_remote_surface *trs, uint32_t type, struct wl_buffer *tbm,
		int32_t img_file_fd, uint32_t img_file_size, uint32_t time, struct wl_array *keys, const char *appid,
		const char *instance_id, int pid, void *data)
{
	Evas_Native_Surface ns;
	tbm_surface_h tbm_surface;
	int width, height;
	char *map;
	char format[] = "png";
	screen_connector_watcher_evas_h watcher_evas_h = (screen_connector_watcher_evas_h)data;
	img_info_h info;
	tbm_surface_h pre_tbm_surface;

	if (tbm == NULL && type == TIZEN_REMOTE_SURFACE_BUFFER_TYPE_TBM) {
		LOGE("TBM is null !! %d, %s, %s", TIZEN_REMOTE_SURFACE_BUFFER_TYPE_TBM, appid, instance_id);
		return;
	}

	info = g_hash_table_lookup(__img_info_table, instance_id);
	if (info == NULL) {
		info = __create_img_info(appid, instance_id, pid);
		g_hash_table_insert(__img_info_table, strdup(instance_id), info);
	}

	/* check type of given buffer */
	if (type == TIZEN_REMOTE_SURFACE_BUFFER_TYPE_TBM) {

		if (info->img_tbm == NULL) {
			LOGD("first added tbm type !!!! %d, %d %s", (info->img_tbm == NULL), type, appid);
			if (info->img_file) {
				watcher_evas_h->ops->removed_cb(appid, instance_id, pid, watcher_evas_h->data);
				__clear_img_file(info);
			}
			info->img_tbm = evas_object_image_filled_add(
					evas_object_evas_get(__viewer_win));
			watcher_evas_h->ops->added_cb(appid, instance_id, pid, watcher_evas_h->data);
		}

		/* get tbm surface from buffer */
		tbm_surface = wl_buffer_get_user_data(tbm);
		width = tbm_surface_get_width(tbm_surface);
		height = tbm_surface_get_height(tbm_surface);

		memset(&ns, 0, sizeof(Evas_Native_Surface));
		ns.version = EVAS_NATIVE_SURFACE_VERSION;
		ns.type = EVAS_NATIVE_SURFACE_TBM;
		ns.data.tbm.buffer = tbm_surface;

		evas_object_resize(info->img_tbm, width, height);
		evas_object_image_size_set(info->img_tbm, width, height);
		evas_object_image_native_surface_set(info->img_tbm, &ns); /* set native surface */
		evas_object_image_pixels_dirty_set(info->img_tbm, EINA_TRUE); /* set dirty for image updating */
		watcher_evas_h->ops->updated_cb(appid, instance_id, pid, info->img_tbm, watcher_evas_h->data);

	} else if (type == TIZEN_REMOTE_SURFACE_BUFFER_TYPE_IMAGE_FILE) {

		if (info->img_file == NULL) {
			LOGD("first added file type !!!! %d %s", type, appid);
			if (info->img_tbm) {
				watcher_evas_h->ops->removed_cb(appid, instance_id, pid, watcher_evas_h->data);
				__clear_img_tbm(info);
			}
			info->img_file = evas_object_image_filled_add(
					evas_object_evas_get(__viewer_win));
			watcher_evas_h->ops->added_cb(appid, instance_id, pid, watcher_evas_h->data);
		}

		map = mmap(NULL, img_file_size, PROT_READ, MAP_SHARED, img_file_fd, 0);
		evas_object_image_memfile_set(info->img_file, map, img_file_size, format, NULL);
		evas_object_image_size_get(info->img_tbm, &width, &height);
		evas_object_image_fill_set(info->img_file, 0, 0, width, height);
		evas_object_resize(info->img_file, width, height);
		munmap(map, img_file_size);

		evas_object_image_native_surface_set(info->img_tbm, NULL); /* set null to previous object for the tbm type */
		watcher_evas_h->ops->updated_cb(appid, instance_id, pid, info->img_file, watcher_evas_h->data);
	}

	if (info->pre_buffer) {
		if (tizen_remote_surface_get_version(trs) >= TIZEN_REMOTE_SURFACE_RELEASE_SINCE_VERSION) {
			tizen_remote_surface_release(trs, info->pre_buffer);
			pre_tbm_surface = (tbm_surface_h)wl_buffer_get_user_data(info->pre_buffer);
			tbm_surface_internal_unref(pre_tbm_surface);
			wl_buffer_destroy(info->pre_buffer);
		}
	}
	info->pre_buffer = tbm;
	close(img_file_fd); /* close passed fd */
}

EXPORT_API screen_connector_watcher_evas_h screen_connector_watcher_evas_add(screen_connector_watcher_evas_ops *ops, void *data)
{
	screen_connector_watcher_evas_h handle;
	screen_connector_watcher_evas_ops *evas_ops;
	screen_connector_watcher_ops *watcher_ops;

	watcher_ops = (screen_connector_watcher_ops *)calloc(1, sizeof(screen_connector_watcher_ops));
	if (watcher_ops == NULL) {
		LOGE("watcher_ops calloc fail");
		return NULL;
	}
	watcher_ops->added_cb = __watcher_added_cb;
	watcher_ops->updated_cb = __watcher_update_cb;
	watcher_ops->removed_cb = __watcher_removed_cb;

	evas_ops = (screen_connector_watcher_evas_ops *)calloc(1, sizeof(screen_connector_watcher_evas_ops));
	if (evas_ops == NULL) {
		free(watcher_ops);
		LOGE("evas_ops calloc fail");
		return NULL;
	}

	memcpy(evas_ops, ops, sizeof(screen_connector_watcher_evas_ops));
	handle = (screen_connector_watcher_evas_h)calloc(1, sizeof(struct _screen_connector_watcher_evas_h));
	if (handle == NULL) {
		LOGE("handle calloc fail");
		free(watcher_ops);
		free(evas_ops);
		return NULL;
	}
	handle->ops = evas_ops;
	handle->data = data;
	handle->watcher_h = screen_connector_watcher_add(watcher_ops, handle);

	return handle;
}

static void __destroy_watcher_evas_handle(screen_connector_watcher_evas_h watcher_evas_h)
{
	if (!watcher_evas_h)
		return;

	if (watcher_evas_h->watcher_h)
		screen_connector_watcher_remove(watcher_evas_h->watcher_h);

	if (watcher_evas_h->ops)
		free(watcher_evas_h->ops);

	free(watcher_evas_h);
}

EXPORT_API int screen_connector_watcher_evas_remove(screen_connector_watcher_evas_h handle)
{
	__destroy_watcher_evas_handle(handle);
	return 0;
}

EXPORT_API int screen_connector_watcher_evas_update(screen_connector_watcher_evas_h watcher_evas_h,
	const char *instance_id)
{
	if (watcher_evas_h == NULL) {
		LOGE("watcher_evas_h is NULL");
		return -1;
	}

	if (watcher_evas_h->watcher_h == NULL) {
		LOGE("watcher_h is NULL");
		return -1;
	}

	return screen_connector_watcher_update(watcher_evas_h->watcher_h, instance_id);
}

EXPORT_API int screen_connector_watcher_evas_set_changed_event_filter(screen_connector_watcher_evas_h watcher_evas_h,
		screen_connector_evas_changed_event_filter_type type)
{
	return screen_connector_watcher_set_changed_event_filter(watcher_evas_h->watcher_h, type);
}
